@@ -15,8 +15,23 @@ SLICED_HASH = 4         # sliced directories from hash of ZOPEID
 
 REPOSITORY = SLICED_HASH
 
-SLICE_WIDTH = 1         # characters per slice
-SLICE_DEPTH = 2         # go this many slices (directories) deep
+PHYSICAL = 0            # SYNC_ZODB uses physical path
+VIRTUAL = 1             # SYNC_ZODB uses path relative to virtual root
+
+ZODB_PATH = VIRTUAL
+
+SLICE_WIDTH = 1         # SLICED* uses this many characters per slice
+SLICE_DEPTH = 2         # SLICED* goes this many slices (directories) deep
+
+# repository file name handling
+
+# %u=user, %p=path, %n=file name, %e=file extension, %c=counter, %t=time
+FILE_FORMAT = "%n%c%e"
+
+DISABLED = 0            # allow file names to begin with 'copy_of_'
+ENABLED = 1             # remove 'copy_of_' prefixes from file names
+
+COPY_OF_PROTECTION = ENABLED
 
 # repository file extension handling
 
@@ -26,11 +41,6 @@ MIMETYPE_REPLACE = 2    # remove extension from ZOPEID, then append mime extensi
 
 REPOSITORY_EXTENSIONS = MIMETYPE_REPLACE
 
-# repository file name format
-
-# %u=user, %p=path, %n=file name, %e=file extension, %c=counter, %t=time
-FILE_FORMAT = "%n%c%e"
-
 # repository undo policy
 
 BACKUP_ON_DELETE = 0    # create a .undo copy only when a file is deleted 
