@@ -1,4 +1,5 @@
 """ ExtFile product module """
+# -*- coding: latin-1 -*-
 ###############################################################################
 #
 # Copyright (c) 2001 Gregor Heine <mac.gregor@gmx.de>. All rights reserved.
@@ -43,34 +44,43 @@ __doc__ = """ExtFile product module.
     The ExtFile-Product works like the Zope File-product, but stores 
     the uploaded file externally in a repository-direcory."""
 
-__version__='$Release: 1.2.0 $'[10:-2]
+__version__='$Release: 1.4.2 $'[10:-2]
 
-#from __main__ import *
-#import Zope
 import Globals
 from Products.ZCatalog.CatalogPathAwareness import CatalogAware
 from OFS.SimpleItem import SimpleItem
 from OFS.PropertyManager import PropertyManager
 from Globals import HTMLFile, MessageDialog, InitializeClass
 from AccessControl import ClassSecurityInfo, getSecurityManager
+from AccessControl import Permissions
 from OFS.content_types import guess_content_type
 from mimetypes import guess_extension
+from webdav.Lockable import ResourceLockedError
 from webdav.common import rfc1123_date
 from DateTime import DateTime
 import urllib, os, string, types, sha, base64
 from os.path import join, isfile
-try: from cStringIO import StringIO
-except: from StringIO import StringIO
+from Products.ExtFile.TM import VTM
+
+from webdav.WriteLockInterface import WriteLockInterface
+from IExtFile import IExtFile
 
 from zLOG import *
 _SUBSYS = 'ExtFile'
-_debug = 1
+_debug = 0
 
-ViewPermission = 'View'
-AccessPermission = 'View management screens'
+try: from zExceptions import Redirect
+except ImportError: Redirect = 'Redirect'
+
+ViewPermission = Permissions.view
+AccessPermission = Permissions.view_management_screens
 ChangePermission = 'Change ExtFile/ExtImage'
 DownloadPermission = 'Download ExtFile/ExtImage'
 
+import re
+copy_of_re = re.compile('(^(copy[0-9]*_of_)+)')
+
+
 from Config import *
 
 bad_chars =  ' ,;()[]{}ݟ'
@@ -93,15 +103,18 @@ def manage_addExtFile(self, id='', title='', descr='', file='',
         id = string.translate(id, TRANSMAP)
     tempExtFile = ExtFile(id, title, descr, permission_check, redirect_default_view)
     self._setObject(id, tempExtFile)
-    self._getOb(id).manage_file_upload(file, content_type)
+    if file != '':
+        self._getOb(id).manage_file_upload(file, content_type)
     if REQUEST is not None:
         return self.manage_main(self, REQUEST, update_menu=0)
 
 
 
-class ExtFile(CatalogAware, SimpleItem, PropertyManager): 
+class ExtFile(CatalogAware, SimpleItem, PropertyManager, VTM):
     """ The ExtFile-Product works like the Zope File-product, but stores 
         the uploaded file externally in a repository-direcory. """
+
+    __implements__ = (IExtFile, WriteLockInterface)
     
     # what properties have we?
     _properties=(
@@ -154,7 +167,7 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
                 {'default':'default.gif'},
             'default':'default.gif'
         }
-    
+
     ################################
     # Init method                  #
     ################################
@@ -208,36 +221,20 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
                         last_mod = long(0)
                     if last_mod > 0 and last_mod < mod_since:
                         # Set headers for Apache caching
-                        #REQUEST.RESPONSE.setHeader('Last-Modified', last_mod)
-                        #REQUEST.RESPONSE.setHeader('Content-Type', self.content_type)
-                        #REQUEST.RESPONSE.setHeader('Content-Length', self.get_size()) # XXX: ???
+                        REQUEST.RESPONSE.setHeader('Last-Modified', last_mod)
+                        REQUEST.RESPONSE.setHeader('Content-Type', self.content_type)
+                        # RFC violation. See http://collector.zope.org/Zope/544
+                        #REQUEST.RESPONSE.setHeader('Content-Length', self.get_size())
                         REQUEST.RESPONSE.setStatus(304)
                         return ''
         
-        #if hasattr(self,'has_preview') and self.has_preview: has_preview = 1
-        #else: has_preview = 0
-        #if not self._access_permitted(REQUEST): preview = 1
-        #if (preview and not has_preview): icon = 1
-        # 
-        #if icon:
-        #    filename = join(Globals.package_home(globals()), self.getIconPath())
-        #    content_type = 'image/gif'
-        #elif preview:
-        #    filename = self._fsname(self.prev_filename)
-        #    content_type = self.prev_content_type
-        #else:
-        #    filename = self._fsname(self.filename)
-        #    content_type = self.content_type
-
         filename, content_type, icon, preview = self._get_file_to_serve(icon, preview)
-        if type(filename) == types.ListType:
-            filename = self._fsname(filename)
+        filename = self._get_fsname(filename)
 
-        if _debug > 1: LOG(_SUBSYS, INFO, '%s, %s, %s, %s' %(filename, content_type, icon, preview))
+        if _debug > 1: LOG(_SUBSYS, INFO, 'serving %s, %s, %s, %s' %(filename, content_type, icon, preview))
 
         cant_read_exc = "Can't read: "
-        if not isfile(filename): self._undo()
-        if isfile(filename):
+        if filename:
             try: size = os.stat(filename)[6]
             except: raise cant_read_exc, ("%s (%s)" %(self.id, filename))
         else:
@@ -248,7 +245,7 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
             icon = 1
 
         if icon==0 and width is not None and height is not None:
-            data = StringIO() # hold resized image
+            data = TemporaryFile() # hold resized image
             try:
                 from PIL import Image
                 im = Image.open(filename) 
@@ -289,7 +286,7 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
     security.declareProtected(ViewPermission, 'view_image_or_file')
     def view_image_or_file(self):
         """ The default view of the contents of the File or Image. """
-        raise 'Redirect', self.absolute_url()
+        raise Redirect, self.absolute_url()
 
     security.declareProtected(ViewPermission, 'link')
     def link(self, text='', **args):
@@ -305,7 +302,7 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
     security.declareProtected(ViewPermission, 'icon_gif')
     def icon_gif(self):
         """ Return an icon for the file's MIME-Type """
-        raise 'Redirect', self._static_url(icon=1)
+        raise Redirect, self._static_url(icon=1)
     
     security.declareProtected(ViewPermission, 'icon_html')
     def icon_html(self):
@@ -315,21 +312,15 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
     security.declareProtected(ViewPermission, 'is_broken')
     def is_broken(self):
         """ Check if external file exists and return true (1) or false (0) """
-        fn = self._fsname(self.filename)
-        if not isfile(fn):
-            self._undo()
-            if not isfile(fn):
-                return 1
-        return 0
+        return not self._get_fsname(self.filename)
     
     security.declareProtected(ViewPermission, 'get_size')
     def get_size(self):
         """ Returns the size of the file or image """
-        fn = self._fsname(self.filename)
-        if not isfile(fn): self._undo()
-        if isfile(fn): size = os.stat(fn)[6]
-        else: size = 0
-        return size
+        fn = self._get_fsname(self.filename)
+        if fn:
+            return os.stat(fn)[6]
+        return 0
     
     # b/w compat
     rawsize = get_size
@@ -342,9 +333,7 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
     
     security.declareProtected(ViewPermission, 'getContentType')
     def getContentType(self):
-        """ Get the content type of a file or image.
-            Returns the content type (MIME type) of a file or image.
-        """
+        """ Returns the content type (MIME type) of a file or image. """
         return self.content_type
         
     security.declareProtected(ViewPermission, 'getIconPath')
@@ -368,28 +357,21 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
     
     security.declareProtected(ViewPermission, 'static_url')
     def static_url(self, icon=0, preview=0):
-        """ Return the static url of the file """
+        """ Returns the static url of the file """
         return self._static_url(icon, preview)
                     
     security.declareProtected(ViewPermission, 'static_mode')
     def static_mode(self):
-        """ Return true if serving static urls """
+        """ Returns true if serving static urls """
         return os.environ.get('EXTFILE_STATIC_PATH') is not None
 
     security.declareProtected(AccessPermission, 'get_filename')
     def get_filename(self):
-        """ Return the file name """
+        """ Returns the filename as file system path. 
+            Used by the ZMI to display the filename.
+        """
         return self._fsname(self.filename)
 
-    security.declareProtected(ViewPermission, 'manage_afterUpdate')
-    def manage_afterUpdate(self, filename, content_type, size):
-        """Called whenever the file data has been updated"""
-        pass
-
-    def _afterUpdate(self):
-        """Called whenever the file data has been updated"""
-        return self.manage_afterUpdate(self.get_filename(), self.content_type, self.get_size())
-        
     ################################
     # Protected management methods #
     ################################
@@ -419,12 +401,10 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
             raise ResourceLockedError, "File is locked via WebDAV"
                             
         if type(file) == types.StringType:
-            from tempfile import TemporaryFile
             temp_file = TemporaryFile()
             temp_file.write(file)
             temp_file.seek(0)
-            file = temp_file
-        return self.manage_file_upload(file, content_type, REQUEST)
+        return self.manage_file_upload(temp_file, content_type, REQUEST)
 
     security.declareProtected(ChangePermission, 'manage_file_upload')
     def manage_file_upload(self, file='', content_type='', REQUEST=None):
@@ -441,9 +421,10 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
         self.content_type = self._get_content_type(file, file.read(100), 
                             self.id, self.content_type)
         file.seek(0)
+        self._register()    # Register with TM
         try:
             new_fn = self._get_ufn(self.filename)
-            self._update_data(file, new_fn)
+            self._update_data(file, self._temp_fsname(new_fn))
         finally: self._dir__unlock()
         self.filename = new_fn
         self._afterUpdate()
@@ -464,9 +445,10 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
         self.content_type = self._get_content_type(file, file.read(100),
                             self.id, self.content_type)
         file.seek(0)
+        self._register()    # Register with TM
         try:
             new_fn = self._get_ufn(self.filename)
-            self._update_data(file, new_fn)
+            self._update_data(file, self._temp_fsname(new_fn))
         finally: self._dir__unlock()
         self.filename = new_fn
         self._afterUpdate()
@@ -485,9 +467,18 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
         self.content_type = self._get_content_type(file, file.read(100),
                             self.id, self.content_type)
         file.seek(0)
+        self._register()    # Register with TM
         try:
-            new_fn = self._get_ufn(self.filename, path=REQUEST.steps[:-2])
-            self._update_data(file, new_fn)
+            # Need to pass in the path as webdav.NullResource calls PUT
+            # on an unwrapped object.
+            try:
+                self.aq_parent # This raises AttributeError if no context
+            except AttributeError:
+                path = self._get_zodb_path(REQUEST.PARENTS[0])
+            else:
+                path = None
+            new_fn = self._get_ufn(self.filename, path=path)
+            self._update_data(file, self._temp_fsname(new_fn))
         finally: self._dir__unlock()
         self.filename = new_fn
         self._afterUpdate()
@@ -535,7 +526,7 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
         try:
             self._copy(infile, outfile)
         except:
-            if UNDO_POLICY==ALWAYS_BACKUP:
+            if isfile(outfile): # This is always a .tmp file
                 try: os.remove(outfile)
                 except OSError: pass
             raise
@@ -544,7 +535,7 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
     
     def _copy(self, infile, outfile):
         """ Read binary data from infile and write it to outfile
-            infile and outfile my be strings, in which case a file with that
+            infile and outfile may be strings, in which case a file with that
             name is opened, or filehandles, in which case they are accessed
             directly.
         """
@@ -552,11 +543,7 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
             try:
                 instream = open(infile, 'rb')
             except IOError:
-                self._undo()
-                try:
-                    instream = open(infile, 'rb')
-                except IOError:
-                    raise IOError, ("%s (%s)" %(self.id, infile))
+                raise IOError, ("%s (%s)" %(self.id, infile))
             close_in = 1
         else:
             instream = infile
@@ -592,10 +579,11 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
         """ Restore filename after delete or copy-paste """
         fn = self._fsname(self.filename)
         if not isfile(fn) and isfile(fn+'.undo'): 
-            os.rename(fn+'.undo', fn)
+            self._register()    # Register with TM
+            os.rename(fn+'.undo', self._temp_fsname(self.filename))
     
     def _fsname(self, filename):
-        """ Generate the full filesystem name, incuding directories from 
+        """ Generates the full filesystem name, incuding directories from 
             self._repository and filename
         """
         path = [INSTANCE_HOME]
@@ -606,9 +594,27 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
             path.append(filename)
         return apply(join, path)
 
+    def _temp_fsname(self, filename):
+        """ Generates the full filesystem name of the temporary file """
+        return '%s.tmp' % self._fsname(filename)
+
+    def _get_fsname(self, filename):
+        """ Returns the full filesystem name, preferring tmp over main. 
+            Also attempts to undo. Returns None if the file is broken.
+        """
+        tmp_fn = self._temp_fsname(filename)
+        if isfile(tmp_fn):
+            return tmp_fn
+        fn = self._fsname(filename)
+        if isfile(fn):
+            return fn
+        self._undo()
+        if isfile(tmp_fn):
+            return tmp_fn
+
     # b/w compat
-    _get_filename = _fsname
-    
+    _get_filename = _get_fsname
+
     def _get_ufn(self, filename, path=None, content_type=None, lock=1):
         """ If no unique filename has been generated, generate one
             otherwise, return the existing one.
@@ -627,15 +633,23 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
                 except OSError: pass
         return new_fn
 
-    def _get_new_ufn(self, path=None, content_type=None, lock=1):
+    def _get_new_ufn(self, path=None, content_type=None, lock=1, copy_of_re=copy_of_re):
         """ Create a new unique filename """
+        id = self.id
+
+        # hack so the files are not named copy_of_foo
+        if COPY_OF_PROTECTION:
+            match = copy_of_re.match(id)
+            if match is not None:
+                id = id[len(match.group(1)):]
+
         # get name and extension components from id
-        pos = string.rfind(self.id, '.')
+        pos = string.rfind(id, '.')
         if (pos+1):
-            id_name = self.id[:pos]
-            id_ext = self.id[pos:]
+            id_name = id[:pos]
+            id_ext = id[pos:]
         else:
-            id_name = self.id
+            id_name = id
             id_ext = ''
 
         if REPOSITORY_EXTENSIONS in (MIMETYPE_APPEND, MIMETYPE_REPLACE):
@@ -643,6 +657,8 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
             if mime_ext is not None:
                 if mime_ext in ('.jpeg', '.jpe'): 
                     mime_ext = '.jpg'   # for IE/Win :-(
+                if mime_ext in ('.obj',):
+                    mime_ext = '.exe'   # b/w compatibility
                 if REPOSITORY_EXTENSIONS == MIMETYPE_APPEND:
                     id_name = id_name + id_ext
                 id_ext = mime_ext
@@ -651,8 +667,7 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
         if path is not None:
             rel_url_list = path
         else:
-            rel_url_list = string.split(self.absolute_url(1), '/')[:-1]
-        rel_url_list = filter(None, rel_url_list)
+            rel_url_list = self._get_zodb_path(self.aq_parent)
 
         dirs = []
         if REPOSITORY == SYNC_ZODB: 
@@ -723,7 +738,7 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
             fn = join(dirpath, string.replace(fileformat, "%c", `counter`))
         else: 
             fn = join(dirpath, string.replace(fileformat, "%c", ''))
-        while isfile(fn) or isfile(fn+'.undo'):
+        while isfile(fn) or isfile(fn+'.undo') or isfile(fn+'.tmp'):
             counter = counter + 1
             fn = join(dirpath, string.replace(fileformat, "%c", `counter`))
         if counter: 
@@ -785,6 +800,14 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
         
         return filename, content_type, icon, preview
 
+    def _get_zodb_path(self, ob):
+        """ Returns the ZODB path of an object """
+        if ZODB_PATH == VIRTUAL:
+            path = ob.absolute_url(1).split('/')
+        else:
+            path = list(ob.getPhysicalPath())
+        return filter(None, path)
+
     def _bytetostring (self, value):
         """ Convert an int-value (file-size in bytes) to an String
             with the file-size in Byte, KB or MB
@@ -805,6 +828,13 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
         strg = strg+typ
         return strg
         
+    def _afterUpdate(self):
+        """ Called whenever the file data has been updated. 
+            Invokes the manage_afterUpdate() hook.
+        """
+        return self.manage_afterUpdate(self._get_fsname(self.filename), 
+                                       self.content_type, self.get_size())
+        
     ################################
     # Special management methods   #
     ################################
@@ -817,20 +847,21 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
         """
         call_afterUpdate = 0
         try: 
-            self.absolute_url(1) # this raises an exception, if no context
-        except: 
-            self._v_has_been_cloned=1   # this is to make webdav COPY work
+            self.aq_parent # This raises AttributeError if no context
+        except AttributeError: 
+            self._v_has_been_cloned=1   # This is to make webdav COPY work
         else:
-            old_fn = self._fsname(self.filename)
-            try:
-                new_fn = new_fn or self._get_new_ufn()
-                if isfile(old_fn):
-                    self._update_data(old_fn, new_fn)
-                self.filename = new_fn
-                call_afterUpdate = 1
-            finally: self._dir__unlock()
-            if call_afterUpdate:
-                self._afterUpdate()
+            fn = self._get_fsname(self.filename)
+            if fn:
+                self._register()    # Register with TM
+                try:
+                    new_fn = new_fn or self._get_new_ufn()
+                    self._update_data(fn, self._temp_fsname(new_fn))
+                    self.filename = new_fn
+                    call_afterUpdate = 1
+                finally: self._dir__unlock()
+                if call_afterUpdate:
+                    self._afterUpdate()
         return ExtFile.inheritedAttribute("manage_afterClone")(self, item)
         
     security.declarePrivate('manage_afterAdd')
@@ -852,16 +883,39 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
         """ This method is called, when the object is deleted. To support 
             undo-functionality and because this happens too, when the object 
             is moved (cut-paste) or renamed, the external file is not deleted. 
-            It are just renamed to filename.undo and remains in the 
+            It is just renamed to filename.undo and remains in the 
             repository, until it is deleted manually.
         """
+        tmp_fn = self._temp_fsname(self.filename)
         fn = self._fsname(self.filename)
-        try:
-            os.rename(fn, fn+'.undo')
-        except OSError:
-            pass
+        if isfile(tmp_fn):
+            try: os.rename(tmp_fn, fn+'.undo')
+            except OSError: pass
+            else:
+                try: os.remove(fn)
+                except OSError: pass
+        elif isfile(fn):
+            try: os.rename(fn, fn+'.undo')
+            except OSError: pass
         return ExtFile.inheritedAttribute("manage_beforeDelete")(self, item, container)
 
+    security.declarePrivate('manage_afterUpdate')
+    def manage_afterUpdate(self, filename, content_type, size):
+        """ This method is called whenever the file data has been updated.
+            May be overridden by subclasses to perform additional operations.
+            The 'filename' argument contains the path as returned by get_fsname().
+        """
+        pass
+
+    security.declarePrivate('get_fsname')
+    def get_fsname(self):
+        """ Returns the current file system path of the file or image. 
+            This path can be used to access the file even while a 
+            transaction is in progress (aka Zagy's revenge :-). 
+            Returns None if the file does not exist in the repository. 
+        """
+        return self._get_fsname(self.filename)
+
     ################################
     # Repository locking methods   #
     ################################
@@ -877,7 +931,44 @@ class ExtFile(CatalogAware, SimpleItem, PropertyManager):
         if hasattr(self, '_v_dir__lock'):
             self._v_dir__lock.release()
             delattr(self, '_v_dir__lock')
-    
+
+    ################################
+    # Transaction manager methods  #
+    ################################
+
+    def _register(self):
+        if _debug: LOG(_SUBSYS, INFO, 'registering %s' % self._v_registered)
+        ExtFile.inheritedAttribute('_register')(self)
+        if _debug: LOG(_SUBSYS, INFO, 'registered %s' % self._v_registered)
+
+    def _begin(self):
+        self._v_begin_called = 1    # for tests
+        if _debug: LOG(_SUBSYS, INFO, 'beginning %s' % self.id) 
+
+    def _finish(self):
+        """ Commits the temporary file """
+        self._v_finish_called = 1   # for tests
+        if self.filename:
+            tmp_fn = self._temp_fsname(self.filename)
+            if _debug: LOG(_SUBSYS, INFO, 'finishing %s' % tmp_fn) 
+            if isfile(tmp_fn):
+                if _debug: LOG(_SUBSYS, INFO, 'isfile %s' % tmp_fn) 
+                fn = self._fsname(self.filename)
+                try: os.remove(fn)
+                except OSError: pass
+                os.rename(tmp_fn, fn)
+
+    def _abort(self):
+        """ Deletes the temporary file """
+        self._v_abort_called = 1    # for tests
+        if self.filename:
+            tmp_fn = self._temp_fsname(self.filename)
+            if _debug: LOG(_SUBSYS, INFO, 'aborting %s' % tmp_fn) 
+            if isfile(tmp_fn):
+                if _debug: LOG(_SUBSYS, INFO, 'isfile %s' % tmp_fn) 
+                try: os.remove(tmp_fn)
+                except OSError: pass
+
 
 InitializeClass(ExtFile)
 
@@ -928,12 +1019,12 @@ class DirLock:
         self._lock = os.path.join(dir, self.lock_name)
         for i in range(self.sleep_times):
             if self.islocked():
-                if _debug: LOG(_SUBSYS, BLATHER, "Waiting for lock '%s'" %self._lock)
+                LOG(_SUBSYS, BLATHER, "Waiting for lock '%s'" %self._lock)
                 time.sleep(self.sleep_secs)
             else:
                 self._mklock()
                 break
         else:
-            if _debug: LOG(_SUBSYS, BLATHER, "Failed to get lock '%s'" %self._lock)
+            LOG(_SUBSYS, BLATHER, "Failed to get lock '%s'" %self._lock)
             raise DirLockError, "Failed to get lock '%s'" %self._lock
 
