@@ -44,16 +44,28 @@ __doc__ = """ExtImage product module.
     uploaded image externally in a repository-direcory. It creates a preview
     of the image (requires PIL)."""
 
-__version__='$Release: 1.2.0 $'[10:-2]
+__version__='$Release: 1.4.2 $'[10:-2]
 
-#from __main__ import *
-#import Zope
 import Globals
 from Products.ExtFile.ExtFile import *
 from Globals import HTMLFile, MessageDialog, InitializeClass
 from AccessControl import ClassSecurityInfo
+from webdav.Lockable import ResourceLockedError
 import urllib, os, string, types
 from os.path import join, isfile
+from tempfile import TemporaryFile
+
+from webdav.WriteLockInterface import WriteLockInterface
+from IExtFile import IExtImage
+
+from zLOG import *
+_SUBSYS = 'ExtImage'
+_debug = 0
+
+try: from zExceptions import Redirect
+except ImportError: Redirect = 'Redirect'
+
+from Config import REPOSITORY_UMASK
 
 NO_PREVIEW = 0
 GENERATE = 1
@@ -77,7 +89,8 @@ def manage_addExtImage(self, id='', title='', descr='', file='', preview='',
         id = string.translate(id, TRANSMAP)
     tempExtImage = ExtImage(id, title, descr, permission_check, redirect_default_view)
     self._setObject(id, tempExtImage)
-    self._getOb(id).manage_file_upload(file, content_type, 0, create_prev, maxx, maxy, ratio)
+    if file != '':
+        self._getOb(id).manage_file_upload(file, content_type, 0, create_prev, maxx, maxy, ratio)
     if create_prev==UPLOAD_NORESIZE or create_prev==UPLOAD_RESIZE:
         self._getOb(id).manage_file_upload(preview, content_type, 1, create_prev, maxx, maxy, ratio)
     if REQUEST is not None:
@@ -90,6 +103,8 @@ class ExtImage(ExtFile):
         uploaded image externally in a repository-direcory. It can create a 
         preview of the image (requires PIL)."""
 
+    __implements__ = (IExtImage, WriteLockInterface)
+
     security = ClassSecurityInfo()
     
     # what do people think they're adding? 
@@ -127,7 +142,7 @@ class ExtImage(ExtFile):
             'height', 'width', and 'alt'. If specified, the 'scale', 'xscale',
             and 'yscale' keyword arguments will be used to automatically adjust
             the output height and width values of the image tag.
-            Adopted and adapted from OFS/image.py
+            Adopted and adapted from OFS/Image.py
         """
         if REQUEST is None and hasattr(self,'REQUEST'): 
             REQUEST = self.REQUEST
@@ -166,7 +181,7 @@ class ExtImage(ExtFile):
     security.declareProtected(ViewPermission, 'preview')
     def preview(self):
         """ Return a preview of the image """
-        raise 'Redirect', self._static_url(preview=1)
+        raise Redirect, self._static_url(preview=1)
     
     security.declareProtected(ViewPermission, 'preview_html')
     def preview_html(self):
@@ -176,12 +191,9 @@ class ExtImage(ExtFile):
     security.declareProtected(ViewPermission, 'is_broken')
     def is_broken(self):
         """ Check if external file exists and return true (1) or false (0) """
-        if self.has_preview and self.filename!=self.prev_filename:
-            fn = self._fsname(self.prev_filename)
-            if not isfile(fn):
-                self._undo()
-                if not isfile(fn):
-                    return 1
+        if self.has_preview and self.filename != self.prev_filename:
+            if not self._get_fsname(self.prev_filename):
+                return 1
         return ExtImage.inheritedAttribute("is_broken")(self)
     
     security.declareProtected(ViewPermission, 'is_webviewable')
@@ -196,11 +208,10 @@ class ExtImage(ExtFile):
     security.declareProtected(ViewPermission, 'prev_rawsize')
     def get_prev_size(self):
         """ Returns the size of the preview file """
-        fn = self._fsname(self.prev_filename)
-        if not isfile(fn): self._undo()
-        if isfile(fn): size = os.stat(fn)[6]
-        else: size = 0
-        return size
+        fn = self._get_fsname(self.prev_filename)
+        if fn: 
+            return os.stat(fn)[6]
+        return 0
 
     # b/w compat
     prev_rawsize = get_prev_size
@@ -235,14 +246,14 @@ class ExtImage(ExtFile):
         """ Get the Fileformat of the image """
         try:
             from PIL import Image
-            im = Image.open(self._fsname(self.filename)) 
+            im = Image.open(self._get_fsname(self.filename)) 
             return im.format
         except:
             return 'unknown'
     
     security.declareProtected(AccessPermission, 'get_prev_filename')
     def get_prev_filename(self):
-        """ Return the preview file name """
+        """ Returns the preview file name for display """
         return self._fsname(self.prev_filename)
         
     ################################
@@ -257,9 +268,17 @@ class ExtImage(ExtFile):
     def manage_del_prev(self, REQUEST=None):
         """ Delete the Preview Image """
         if self.has_preview and self.filename != self.prev_filename:
-            prev_fn = self._fsname(self.prev_filename)
-            try: os.rename(prev_fn,prev_fn+'.undo')
-            except OSError: pass
+            tmp_fn = self._temp_fsname(self.prev_filename)
+            fn = self._fsname(self.prev_filename)
+            if isfile(tmp_fn):
+                try: os.rename(tmp_fn, fn+'.undo')
+                except OSError: pass
+                else:
+                    try: os.remove(fn)
+                    except OSError: pass
+            elif isfile(fn):
+                try: os.rename(fn, fn+'.undo')
+                except OSError: pass
         self.prev_content_type = ''
         self.has_preview = 0
         if REQUEST is not None:
@@ -270,6 +289,7 @@ class ExtImage(ExtFile):
         """ Create a preview Image """
         maxx, maxy = self._formatDimensions(maxx, maxy)
         if maxx!=0 and maxy!=0:
+            self._register()    # Register with TM
             try:
                 new_fn = self._get_ufn(self.prev_filename, content_type='image/jpeg')
                 self._createPreview(self.filename, new_fn, maxx, maxy, ratio)
@@ -303,12 +323,10 @@ class ExtImage(ExtFile):
             raise ResourceLockedError, "File is locked via WebDAV"
 
         if type(file) == types.StringType:
-            from tempfile import TemporaryFile
             temp_file = TemporaryFile()
             temp_file.write(file)
             temp_file.seek(0)
-            file = temp_file
-        return self.manage_file_upload(file, content_type, is_preview,
+        return self.manage_file_upload(temp_file, content_type, is_preview,
                                        create_prev, maxx, maxy, ratio, REQUEST)
 
     security.declareProtected(ChangePermission, 'manage_file_upload')
@@ -326,9 +344,10 @@ class ExtImage(ExtFile):
                 except: raise cant_read_exc, file
             maxx, maxy = self._formatDimensions(maxx, maxy)
             if create_prev==UPLOAD_RESIZE and maxx!=0 and maxy!=0:
+                self._register()    # Register with TM
                 try:
                     new_fn = self._get_ufn(self.prev_filename, content_type='image/jpeg')
-                    self._update_data(file, new_fn)
+                    self._update_data(file, self._temp_fsname(new_fn))
                 finally: self._dir__unlock()
                 self._createPreview(new_fn, new_fn, maxx, maxy, ratio)
             else:
@@ -337,9 +356,10 @@ class ExtImage(ExtFile):
                 self.prev_content_type = self._get_content_type(file, file.read(100), 
                                          self.id, self.prev_content_type)
                 file.seek(0)
+                self._register()    # Register with TM
                 try:
                     new_fn = self._get_ufn(self.prev_filename, content_type=self.prev_content_type)
-                    self._update_data(file, new_fn)
+                    self._update_data(file, self._temp_fsname(new_fn))
                 finally: self._dir__unlock()
                 self.prev_filename = new_fn
                 self._initPreview()
@@ -348,6 +368,7 @@ class ExtImage(ExtFile):
             if create_prev==GENERATE:
                 maxx, maxy = self._formatDimensions(maxx, maxy)
                 if maxx!=0 and maxy!=0:
+                    self._register()    # Register with TM
                     try:
                         new_fn = self._get_ufn(self.prev_filename, content_type='image/jpeg')
                         self._createPreview(self.filename, new_fn, maxx, maxy, ratio)
@@ -370,9 +391,10 @@ class ExtImage(ExtFile):
             self.prev_content_type = self._get_content_type(file, file.read(100), 
                                      self.id, self.prev_content_type)
             file.seek(0)
+            self._register()    # Register with TM
             try:
                 new_fn = self._get_ufn(self.prev_filename, content_type=self.prev_content_type)
-                self._update_data(file, new_fn)
+                self._update_data(file, self._temp_fsname(new_fn))
             finally: self._dir__unlock()
             self.prev_filename = new_fn
             self._initPreview()
@@ -380,6 +402,7 @@ class ExtImage(ExtFile):
             ExtImage.inheritedAttribute("manage_http_upload")(self, url)
             if self.has_preview:
                 width, height = self._getImageSize(self.prev_filename)
+                self._register()    # Register with TM
                 try:
                     new_fn = self._get_ufn(self.prev_filename, content_type='image/jpeg')
                     self._createPreview(self.filename, new_fn, width, height, self.prev_ratio)
@@ -393,8 +416,17 @@ class ExtImage(ExtFile):
         RESPONSE = ExtImage.inheritedAttribute("PUT")(self, REQUEST, RESPONSE)
         if self.has_preview:
             width, height = self._getImageSize(self.prev_filename)
+            self._register()    # Register with TM
             try:
-                new_fn = self._get_ufn(self.prev_filename, content_type='image/jpeg', path=REQUEST.steps[:-2])
+                # Need to pass in the path as webdav.NullResource calls PUT
+                # on an unwrapped object.
+                try:
+                    self.aq_parent # This raises AttributeError if no context
+                except AttributError:
+                    path = self._get_zodb_path(REQUEST.PARENTS[0])
+                else:
+                    path = None
+                new_fn = self._get_ufn(self.prev_filename, content_type='image/jpeg', path=path)
                 self._createPreview(self.filename, new_fn, width, height, self.prev_ratio)
             finally: self._dir__unlock() 
         return RESPONSE
@@ -407,38 +439,45 @@ class ExtImage(ExtFile):
         """ Return width, height tuple using PIL """
         try:
             from PIL import Image
-            try: 
-                im = Image.open(self._fsname(filename)) 
-            except:
-                self._undo()
-                im = Image.open(self._fsname(filename)) 
-                raise
-            width = im.size[0]
-            height = im.size[1]
+            im = Image.open(self._get_fsname(filename)) 
+            return im.size[0], im.size[1]
         except:
-            width = height = 0
-        return width, height
+            return 0, 0
     
     def _createPreview(self, from_filename, to_filename, maxx, maxy, ratio):
         """ Generate a preview using PIL """
         try:
             from PIL import Image
-            im = Image.open(self._fsname(from_filename)) 
-            if im.mode!='RGB' and im.mode!='CMYK': 
-                im = im.convert('RGB')
-            filter = Image.BICUBIC
-            if hasattr(Image, 'ANTIALIAS'): # PIL 1.1.3
-                filter = Image.ANTIALIAS
-            if ratio:               # keep aspect-ratio
-                im.thumbnail((maxx,maxy), filter)
-            else:                   # distort to fixed size
-                im = im.resize((maxx,maxy), filter)
-            im.save(self._fsname(to_filename), 'JPEG', quality=85)
-            self.prev_content_type = 'image/jpeg'
-            self.prev_filename = to_filename
-            self.prev_ratio = ratio
-        except:
+        except ImportError:
             pass
+        else:
+            imfile = self._get_fsname(from_filename)
+            if imfile:
+                im = Image.open(imfile) 
+                if im.mode!='RGB' and im.mode!='CMYK': 
+                    im = im.convert('RGB')
+                filter = Image.BICUBIC
+                if hasattr(Image, 'ANTIALIAS'): # PIL 1.1.3
+                    filter = Image.ANTIALIAS
+                if ratio:               # keep aspect-ratio
+                    im.thumbnail((maxx,maxy), filter)
+                else:                   # distort to fixed size
+                    im = im.resize((maxx,maxy), filter)
+                umask = os.umask(REPOSITORY_UMASK)
+                outfile = self._temp_fsname(to_filename)
+                try:
+                    im.save(outfile, 'JPEG', quality=85)
+                except:
+                    os.umask(umask)
+                    if isfile(outfile):
+                        try: os.remove(outfile)
+                        except OSError: pass
+                    raise
+                else:
+                    os.umask(umask)
+                self.prev_content_type = 'image/jpeg'
+                self.prev_filename = to_filename
+                self.prev_ratio = ratio
         self._initPreview()
     
     def _initPreview(self):
@@ -466,10 +505,11 @@ class ExtImage(ExtFile):
     
     def _undo(self):
         """ Restore filename after delete or copy-paste """
-        if self.has_preview:
+        if self.has_preview and self.filename != self.prev_filename:
             fn = self._fsname(self.prev_filename)
             if not isfile(fn) and isfile(fn+'.undo'): 
-                os.rename(fn+'.undo', fn)           # rename preview
+                self._register()    # Register with TM
+                os.rename(fn+'.undo', self._temp_fsname(self.prev_filename))
         return ExtImage.inheritedAttribute("_undo")(self)
     
     def _get_content_type(self, file, body, id, content_type=None):
@@ -494,26 +534,26 @@ class ExtImage(ExtFile):
             file is created and self.filename is changed.
         """
         try: 
-            self.absolute_url(1)    # this raises an exception, if no context
-        except: 
+            self.aq_parent # This raises AttributeError if no context
+        except AttributeError: 
             pass
         else:
             result = ExtImage.inheritedAttribute("manage_afterClone")(self, item)
+            self._register()    # Register with TM
             try:
                 new_prev_fn = self._get_new_ufn(content_type=self.prev_content_type)
                 if self.has_preview and self.filename != self.prev_filename:
-                    old_prev_fn = self._fsname(self.prev_filename)
-                    if isfile(old_prev_fn):
-                        self._update_data(old_prev_fn, new_prev_fn)
+                    old_prev_fn = self._get_fsname(self.prev_filename)
+                    if old_prev_fn:
+                        self._update_data(old_prev_fn, self._temp_fsname(new_prev_fn))
                         self.prev_filename = new_prev_fn
                     else:
                         self.prev_filename = []
                         self.has_preview = 0
                 elif self.has_preview:
-                    old_fn = self._fsname(self.filename)
-                    if isfile(old_fn) or isfile(old_fn+'.undo'): 
-                        self.prev_filename = new_prev_fn
-                    else:
+                    # XXX: This seems to be an impossible state?
+                    old_fn = self._get_fsname(self.filename)
+                    if not old_fn:
                         self.prev_filename = []
                         self.has_preview = 0
                 else:
@@ -536,17 +576,51 @@ class ExtImage(ExtFile):
         """ This method is called, when the object is deleted. To support 
             undo-functionality and because this happens too, when the object 
             is moved (cut-paste) or renamed, the external file is not deleted. 
-            It are just renamed to filename.undo and remains in the 
+            It is just renamed to filename.undo and remains in the 
             repository, until it is deleted manually.
         """
         if self.has_preview and self.filename != self.prev_filename:
-            prev_fn = self._fsname(self.prev_filename)
-            try:                                
-                os.rename(prev_fn,prev_fn+'.undo')
-            except OSError:
-                pass
+            tmp_fn = self._temp_fsname(self.prev_filename)
+            fn = self._fsname(self.prev_filename)
+            if isfile(tmp_fn):
+                try: os.rename(tmp_fn, fn+'.undo')
+                except OSError: pass
+                else:
+                    try: os.remove(fn)
+                    except OSError: pass
+            elif isfile(fn):
+                try: os.rename(fn, fn+'.undo')
+                except OSError: pass
         return ExtImage.inheritedAttribute("manage_beforeDelete")(self, item, container)
 
+    ################################
+    # Transaction manager methods  #
+    ################################
+
+    def _finish(self):
+        """ Commits the temporary file """
+        if self.prev_filename and self.filename != self.prev_filename:
+            tmp_fn = self._temp_fsname(self.prev_filename)
+            if _debug: LOG(_SUBSYS, INFO, 'finishing %s' % tmp_fn)
+            if isfile(tmp_fn):
+                if _debug: LOG(_SUBSYS, INFO, 'isfile %s' % tmp_fn)
+                fn = self._fsname(self.prev_filename)
+                try: os.remove(fn)
+                except OSError: pass
+                os.rename(tmp_fn, fn)
+        ExtImage.inheritedAttribute('_finish')(self)
+    
+    def _abort(self):
+        """ Deletes the temporary file """
+        if self.prev_filename and self.filename != self.prev_filename:
+            tmp_fn = self._temp_fsname(self.prev_filename)
+            if _debug: LOG(_SUBSYS, INFO, 'aborting %s' % tmp_fn)
+            if isfile(tmp_fn):
+                if _debug: LOG(_SUBSYS, INFO, 'isfile %s' % tmp_fn)
+                try: os.remove(tmp_fn)
+                except OSError: pass
+        ExtImage.inheritedAttribute('_abort')(self)
+        
 
 InitializeClass(ExtImage)
 
